-- Koji Kagawa
-- Parser for Polymorphic Variant

module Parser where {

import Maybe (fromJust);
import Data.Char (isUpper, isLower);

import Text.ParserCombinators.Parsec;
import Text.ParserCombinators.Parsec.Char;
import Text.ParserCombinators.Parsec.Expr;
import qualified Text.ParserCombinators.Parsec.Token as P;
import Text.ParserCombinators.Parsec.Language (haskellStyle);

import PPrint (pprint, render);
import Id;
import Kind;
import Type;
import Subst;
import Pred;
import Scheme;
import Assump;
import Pat;
import Lit;
import TIMonad;
import TIMain;
import StaticPrelude;
import KindInfer;
import List (nub);

-- for debug
-- import Hugs.Observe;

type InstDefn = ((Id{- record name -}, Id{- variant name -}), [(Id, Maybe Scheme, [Alt])]);

type MyState = (ClassEnv, [InstDefn], [Assump]);

type MyParser a = GenParser Char MyState a;

-- MyStatesetter, getter
modifyClassEnv id c = updateState (\ (ce, ins, as) -> (modify ce id c, ins, as));

getClassEnv = do { (ce, _, _) <- getState; return ce };

setClassEnv ce = do { (_, ins, as) <- getState; setState (ce, ins, as) };

modifyAssump id scm = updateState (\ (ce, ins, as) -> (ce, ins, (id:>:scm):as));

getAssump = do { (_, _, as) <- getState; return as };

setAssump as = do { (ce, ins, _) <- getState; setState (ce, ins, as) };

addInstDefn vid rid decls  = updateState (\ (ce, ins, as) -> (ce, ((vid,rid),decls):ins, as));

getInstDefn = do { (_, ins, _) <- getState; return ins };

-- setInstDefn ins = do { (ce, _, as) <- getState; setState (ce, ins, as) };

-- parse as the Raw format before inferring Kinds
data RawTopDecl = RawVariantDecl [RawPred] RawPred [Decl]
 	        | RawRecordDecl  [RawPred] RawPred [Decl]
		| RawInstDecl RawInst [Decl] 
                | RawTopDecl Decl 
	       deriving Show
;

-- Raw format for Declaration
data Decl = GenDecl Id ([RawPred], RawType)
	  | Decl Id Alt
	  | EmptyDecl  
	    deriving Show
;

-- aux function that collects definitions of a function
declsSplit v [] = ([], []);
declsSplit v (ds0@(Decl v1 alt : ds1)) =
	   if v==v1 then let { (as, ds2) = declsSplit v ds1 } in (alt:as, ds2)
                    else ([], ds0);	       
declsSplit v ds1 = ([], ds1);


declsCollect :: Either [Tyvar] ClassEnv -> [Decl] -> [(Id,Maybe Scheme,[Alt])];
declsCollect vs [] = [];
declsCollect vs (EmptyDecl:ds) = declsCollect vs ds;
declsCollect vs (GenDecl v t : ds) =
    let { (as, ds1) = declsSplit v ds;
	  t1 = case vs of
	  {
	   -- Left means we already know the kind of the variable
	   Left vs0 -> kindReplaceQual vs0 t;
	   -- Right means we do not yet know ...
	   Right ce ->
	   let { (ps, e) = t;
		 tvs = varsExpr e (concat (map snd ps)); -- variables in t
		 vs0 =
		 runTI $
		 do {
		     kindExp ce t;
		     s <- getSubst;
		     let { f id = Tyvar id (type2kind (apply s (TVar (Tyvar id Star)))) };
		     return (map f tvs)
		    }
	       } in kindReplaceQual vs0 t
	   }
	} in
    (v, Just (quantify (tv t1) t1), as) : declsCollect vs ds1;
declsCollect vs (Decl v alt : ds) =
    let { (as, ds1) = declsSplit v ds } in
    (v, Nothing, alt:as) : declsCollect vs ds1;

decls2Bg :: Either [Tyvar] ClassEnv -> [Decl] -> BindGroup;
decls2Bg vs ds = toBg (declsCollect vs ds);

----------------------------------------------------------------------
-- the parser

myLexer = P.makeTokenParser haskellStyle 
                { P.reservedOpNames = ["::","=","\\","->","@","=>"]
		, P.identStart     = letter <|> char '_'
                , P.reservedNames  =
		  ["let","in","case","of","if","then","else",
                   {- "data",-} "type",
                   {- "class", -} "record", "variant",
                   "infix","infixl","infixr",
		   "instance",
                   "newtype","where"
                  ]
                }
;

parens          = P.parens myLexer;    
braces          = P.braces myLexer;    
brackets        = P.brackets myLexer;    
semiSep1        = P.semiSep1 myLexer;    
semiSep         = P.semiSep myLexer;    
commaSep1       = P.commaSep1 myLexer;    
commaSep        = P.commaSep myLexer;    
whiteSpace      = P.whiteSpace myLexer;    
symbol          = P.symbol myLexer;    
identifier      = P.identifier myLexer;    
reserved        = P.reserved myLexer;    
reservedOp      = P.reservedOp myLexer;    
float           = P.float myLexer;
integer         = P.integer myLexer;
natural         = P.natural myLexer;    
charLiteral     = P.charLiteral myLexer;    
stringLiteral   = P.stringLiteral myLexer;    

{-
topdecls -> topdecl1 ; ... ; topdecln
-}

parseRawTopDecls :: MyParser [RawTopDecl];
parseRawTopDecls = do {
		       r <- semiSep1 parseRawTopDecl;
		       eof;
		       return r;
		   };
		
{-
topdecl -> record [scontext =>] tycls tyvar [where cdecls]
	|  variant [scontext =>] tycls tyvar [where cdecls]
	|  instance [scontext =>] tycls inst [where idecls]
	|  decl
-}

parseRawTopDecl = parseRawRecordDecl
              <|> parseRawVariantDecl
	      <|> parseRawInstanceDecl
	      <|> do { d <- parseDecl; return (RawTopDecl d) }
;

conid = try (do {
		 xs <- identifier;
		 if isUpper (head xs) then return xs else fail "Uppercase"
		})
;

varid = try (do {
		 xs <- identifier;
		 let { c = head xs } ;
		 if isLower c || c =='_' then return xs else fail "Lowercase"
		})
;

parseRawClassDecl = do {
		      sc <- parseSContext;
		      tv <- varid;
		      reservedOp "::";
		      (c, targs) <- parseInst;
		      reserved "where";
		      cdecls <- parseCDecls;
		      return (sc, (c, tv:targs), cdecls) 
		    }
;

parseRawRecordDecl = do {
		       reserved "record";
		       (sc, p, es) <- parseRawClassDecl;
		       return (RawRecordDecl sc p es)
                     }
;

parseRawVariantDecl = do {
		       reserved "variant";
		       (sc, p, es) <- parseRawClassDecl;
		       return (RawVariantDecl sc p es)
                     }
;

parseRawInstanceDecl = do {
			 reserved "instance";
			 sc <- parseSContext;
			 (v, vs) <- parseInst;
			 reservedOp "::";
			 x <- varid;
			 reservedOp "::";			
			 (r, rs) <- parseInst;
			 reserved "where";
			 ids <- parseIDecls;
			 return (RawInstDecl (sc, ((v, vs),x, (r, rs))) ids)
		       }
;	   

{-
scontext -> simpleclass
	| ( simpleclass1 , ... , simpleclassn ) 	(n>=0)
simpleclass 	-> tycls tyvar
tycls  -> conid
tyvar  -> varid

-}

parseSContext :: MyParser [RawPred];
parseSContext = try (do { sc <- parseSimpleClass;
			  reservedOp "=>";
			  return [sc]
			})
		<|> try (do { xs <- parens (commaSep1 parseSimpleClass);
			  reservedOp "=>";
			  return xs
			})
		<|> return []
;

parseSimpleClass = do {
		       v <- varid;
		       reservedOp "::";
		       (c, targs) <- parseInst;
		       return (c, v:targs)
		      }
;


{-
inst 	 -> 	 gtycon tyvar1 ... tyvark  	(k>=0, tyvars distinct)
gtycon 	 -> 	 tycon
-}

parseInst :: MyParser (Id, [Id]);
parseInst = do { i <- parseGTycon;
		 vs <- sepBy varid whiteSpace;
		 return (i, vs)
	       }
;

parseGTycon = conid 
           {- <|> ... -}
;  


{-
cdecls 	-> 	 { cdecl1 ; ... ; cdecln } 	 (n>=0)
cdecl 	-> 	gendecl
	| 	(funlhs | var) rhs
gendecl -> var :: [context =>] type
funlhs ->  var apat {apat}
	|  funlhs apat {apat}

-}

parseCDecls = braces (semiSep parseCDecl);

parseCDecl :: MyParser Decl;
parseCDecl  = (do {
		  id <- identifier;
		  (do {
		       reservedOp "::";
		       c <- parseSContext;
		       t <- parseType;
		       return (GenDecl id (c, t))  
		      })
		  <|>
		  (do {
		       pats <- many parseAPat;
		       reservedOp "=";
		       r <- parseExp;
		       return (Decl id (pats, r))
		      })
		 })
	      <|>
	      (do {
		   (vid, pats1) <- parens parseFunLhs;
		   pats2 <- many parseAPat;
		   reservedOp "=";
		   r <- parseExp;
		   return (Decl vid (pats1++pats2, r))
		   }) <?> "parseCDecl"
;


parseInstFunLhs = (do {
		    id <- varid;
		    pats <- parseIPats;
		    return (id, pats)
		   })
	      	  <|> 
	      	  (do {
	      	  	   (id, pats1) <- parens parseInstFunLhs;
	      	  	   pats2 <- many parseAPat;
	      	  	   return (id, pats1++pats2)
	      	  	  })
	      	  <?> "parseInstFunLhs"
;


parseFunLhs = (do {
		  id <- varid;
		  pats <- many parseAPat;
		  return (id, pats)
		 })
	      <|>
	      (do {
		   (id, pats1) <- parens parseFunLhs;
		   pats2 <- many parseAPat;
		   return (id, pats1++pats2)
		  })
	      <?> "parseFunLhs"
;

{-
idecls	-> 	{ idecl1 ; ... ; idecln } 	(n>=0)
idecl 	-> 	(funlhs | var) rhs
	|       (empty)
 -}

-- N.B. only top level varid of the first parameter is put into PCon.
parseIPat = do {
              v <- varid;
  	      return (PCon v [])
	    }
            <|>
            parens (do {
                       v <- varid;
                       apats <- many parseAPat;
                       return (PCon v apats)
                    }) 
	    <?> "parseIPat"
;

parseIPats = do {
	       p  <- parseIPat;
               ps <- many parseAPat;
               return (p:ps)
             }
;


parseIDecls = braces (semiSep parseIDecl);
parseIDecl :: MyParser Decl;
parseIDecl  = (do {
		  id <- varid;
		  pats <- parseIPats;
		  reservedOp "=";
		  r <- parseExp;
		  return (Decl id (pats, r))
		  })
	      <|>
	      (do {
		   (vid, pats1) <- parens parseInstFunLhs;
		   pats2 <- many parseAPat;
		   reservedOp "=";
		   r <- parseExp;
		   return (Decl vid (pats1++pats2, r))
		   })
	      <|>
	      return EmptyDecl
	      <?> "parseIDecl"	     
;
{-
decl -> gendecl
      | funlhs rhs
rhs 	->  = exp 
-}

parseGenDecl = (do {
		    v <- varid;
		    reservedOp "::";
		    c <- parseSContext;
		    t <- parseType;
		    return (GenDecl v (c, t))
		    })
	       <|> return EmptyDecl
;

parseDecl :: MyParser Decl; 
parseDecl =
    try (do {
	 g <- try parseGenDecl;
	 return g
	})
    <|>
    (do {
	 (id, pats) <- parseFunLhs; 
	 reserved "=";
	 e <- parseExp;
	 return (Decl id (pats, e))
	})
    <?> "parseDecl"
;

parseDecls :: MyParser [Decl]; 
parseDecls = braces (semiSep parseDecl);

----------------------------------------------------------------------
	 
parsePat :: MyParser Pat;
parsePat = buildExpressionParser patTable parsePat10 <?> "pattern"
;

patTable = [[ Infix (do { try (symbol ":"); return pCons}) AssocRight]]
;

parsePat10 = try (do {
		       c <- conid;
		       apats <- many parseAPat;
		       return (PCon c apats)
		      })
	     <|>
	     parseAPat
	     <?> "parsePat10"
;

tupCfun n =
	[tup2Cfun, tup3Cfun, tup4Cfun, tup5Cfun, tup6Cfun, tup7Cfun] !! (n-2);

parseAPat :: MyParser Pat;
parseAPat = (do { v <- varid;
		  (do {
		       reservedOp "@";
		       a <- parseAPat;
		       return (PAs v a)
		       }
		   <|>
		   return (if v == "_" then PWildcard else PVar v))
		 })
	    <|>
	    (do { c <- conid;
		  -- as <- getAssump;
		  -- ts <- find c as;
		  -- return (PCon (c :>: ts) []) 
                  return (PVar c)
		})
	    <|>
	    (do { l <- parseLiteral; return (PLit l) })
	    <|> parens (do { pats <- commaSep parsePat;
			     let { len = length pats };
			     return 
			     (if len==0 then (PCon (idAssump unitCfun) [])
			      else if len==1 then head pats
			      else (PCon (idAssump (tupCfun len)) pats))
			   })
	    <?> "aPat"
;

parseExp = buildExpressionParser expTable parseExp10 <?> "expression";

expTable =
   let { op1 name = \ x -> Ap (Ap (Var name) (Lit (LitInt 0))) x;  
	 op2 name = \ x y -> Ap (Ap (Var name) x) y } in
   [
	 [ Infix (do { try (symbol "."); return (op2 ".")}) AssocRight
	 ],
	 [ Infix (do { try (symbol "^"); return (op2 "^")}) AssocRight,
	   Infix (do { try (symbol "^^"); return (op2 "^^")}) AssocRight,
	   Infix (do { try (symbol "**"); return (op2 "**")}) AssocRight
	 ],
	 [ Infix (do { try (symbol "*"); return (op2 "*")}) AssocLeft,
	   Infix (do { try (symbol "/"); return (op2 "/")}) AssocLeft,
	   Infix (do { try (symbol "`quot`"); return (op2 "quot")}) AssocLeft,
	   Infix (do { try (symbol "`rem`"); return (op2 "rem")}) AssocLeft,
	   Infix (do { try (symbol "`div`"); return (op2 "div")}) AssocLeft,
	   Infix (do { try (symbol "`mod`"); return (op2 "mod")}) AssocLeft
	 ],
	 [
	   Infix (do { try (symbol "+"); return (op2 "+")}) AssocLeft,
	   Infix (do { try (symbol "-"); return (op2 "-")}) AssocLeft,
	   Prefix (do { try  (symbol "-"); return (op1 "-")}),
	   Infix (do { try (symbol "+."); return (op2 "+.")}) AssocLeft,
	   Infix (do { try (symbol "-."); return (op2 "-.")}) AssocLeft
	 ],
	 [ Infix (do { try (symbol "++"); return (op2 "++")}) AssocLeft ],
	 (map (\ op -> Infix (do { try (symbol op); return (op2 op)})
	                     AssocNone)
	  ["==", "/=", "<", "<=", ">=", ">"]),
	 [ Infix (do { try (symbol "&&");
		       return (\ x y -> If x y (Lit (LitBool False))) })
	   AssocRight
	 ],
	 [ Infix (do { try (symbol "||");
		       return (\ x y -> If x (Lit (LitBool True)) y) })
	   AssocRight
	 ],
	 [ Infix (do { try (symbol ">>"); return (op2 ">>")}) AssocLeft,
	   Infix (do { try (symbol ">>="); return (op2 ">>=")}) AssocLeft,
	   Infix (do { try (symbol "=<<"); return (op2 "=<<")}) AssocRight
	 ],
	 [ Infix (do { try (symbol "$"); return (op2 "$")}) AssocRight,
	   Infix (do { try (symbol "$!"); return (op2 "$!")}) AssocRight,
	   Infix (do { try (symbol "`seq`"); return (op2 "seq")}) AssocRight
	 ]
]
;

parseExp10 :: MyParser Expr;
parseExp10 = do {
		 reservedOp "\\";
		 pats <- many1 parseAPat;
		 reservedOp "->";
		 e <- parseExp;
		 return (Lam (pats, e))
		 }
	     <|>
	     do {
		 reserved "let";
		 decls <- parseDecls;
		 reserved "in";
		 e <- parseExp;
		 ce <- getClassEnv;
		 return (Let (decls2Bg (Right ce) decls) e)
		 } 
             <|> 
	     do {
		 reserved "if";
		 e1 <- parseExp;
		 reserved "then";
		 e2 <- parseExp;
		 reserved "else";
		 e3 <- parseExp;
		 return (If e1 e2 e3)
		}
             <|>
	     do {
		 reserved "case";
		 e <- parseExp;
		 reserved "of";
		 alts <- braces (semiSep1 parseAlt);
		 return (Case e alts)
		}
	     {- <|>
		do { stmts } -}
	     <|>
	     parseFExp
;

parseFExp :: MyParser Expr;
parseFExp =  do {
		 es <- many1 parseAExp;
		 return (foldl1 Ap es)
		} <?> "parseFExp"
;

parseAExp :: MyParser Expr;
parseAExp = do {
		v <- varid;
		return (Var v)
	       }
	    <|>
	    do {
		c <- conid;
                return (Const c)
	       }
            <|>
	    do {
		l <- parseLiteral;
		return (Lit l)
	       }
            <|>
	    do {
		es <- parens (commaSep parseExp);
		let { l = length es };
		return (if l == 0 then Const (idAssump unitCfun)
			else if l == 1 then head es
			else foldl Ap (Const (idAssump (tupCfun l))) es)
	       }
            <|>
	    do {
		es <- brackets (commaSep parseExp);
		return (foldr (\ a as -> Ap (Ap (Const (idAssump consCfun)) a) as)
			      (Const (idAssump nilCfun)) es)
	       }
    <?> "parseAExp"
;

parseLiteral :: MyParser Literal;
parseLiteral = try (
	       do {
		n <- natural;
		return (LitInt n)
	       })
	       <|>
	       do {
		   f <- float;
		   return (LitRat (toRational f))
		  }
               <|>
	       do {
		   s <- stringLiteral;
		   return (LitStr s)
		   }
	       <|>
	       do {
		   c <- charLiteral;
		   return (LitChar c)
		   }
;


parseAlt :: MyParser (Pat, Expr);
parseAlt = do {
	       pat <- parsePat;
	       reservedOp "->";
	       e <- parseExp;
	       return (pat, e)
	      }
;

parseType :: MyParser Expr;
parseType = let { fne f a = Ap (Ap (tCon2Const tArrow) f) a } in  
            do {
		ts <- sepBy1 parseBType (reservedOp "->");
		return (foldr1 fne ts)
	       }
;

parseBType :: MyParser Expr;
parseBType = do {
		 ts <- many1 parseAType;
		 return (foldl1 Ap ts)
		}
;


tCon2Const (TCon (Tycon str kind)) = Const (str {- :>: Forall [] ([] :=> (kind2type kind)) -});

tTupleN n = tCon2Const 
	    ([tUnit, error "tTupleN", tTuple2, tTuple3, tTuple4, tTuple5, tTuple6, tTuple7] !! n)
	    ;


findTCon ce c = case lookup c [("Bool", tBool)] of
	     {
	      Just t -> t;
	      Nothing -> let { tv:tvs =  map (\ (Tyvar _ k) -> k) (sig ce c) } in
	                 TCon (Tycon c (foldr Kfun tv tvs))
 	     }
;

parseAType :: MyParser Expr;
parseAType = do {
		 c <- conid;
		 ce <- getClassEnv;
		 return (tCon2Const (findTCon ce c))
		}
             <|>
	     do {
		 v <- varid;
		 return (Var v);
		}
    	     <|>
	     do {
		 ts <- parens (commaSep parseType);
		 let { len = length ts };
		 return (case len of {
				      0 -> tTupleN 0;
				      1 -> head ts;
				      _ -> foldl Ap (tTupleN len) ts
           			     })
		 }
             <|>
	     do {
		 t <- brackets parseType;
		 return {- (TAp tList t) -} (Ap (tCon2Const tList) t)
		}
;
}
