-----------------------------------------------------------------------------
-- Pred:		Predicates
-- 
-- Part of `Typing Haskell in Haskell', version of November 23, 2000
-- Copyright (c) Mark P Jones and the Oregon Graduate Institute
-- of Science and Technology, 1999-2000
-- 
-- This program is distributed as Free Software under the terms
-- in the file "License" that is included in the distribution
-- of this software, copies of which may be obtained from:
--             http://www.cse.ogi.edu/~mpj/thih/
-- 
-----------------------------------------------------------------------------

module Pred where
import List(union,(\\),partition,find,nub)
import Monad(msum)
import Id
import Kind
import Type
import Subst
import Unify
import PPrint

-- for debug
-- import Hugs.Observe

data Qual t = [Pred] :=> t
              deriving (Show, Eq)  -- Show added by kagawa

instance PPrint t => PPrint (Qual t) where
  pprint (ps  :=> t) = (pprint ps <+> text ":=>") $$ nest 2 (parPprint t)

data Pred   = IsIn Id [Type]
              deriving (Show, Eq)  -- Show added by kagawa

predHead            :: Pred -> Id
predHead (IsIn i ts) = i

instance PPrint Pred where
  pprint (IsIn i [t]) = text "isIn1" <+> text ("c" ++ i) <+> parPprint t
  pprint (IsIn i ts)  = text "isIn" <+> text ("c" ++ i) <+> pplist ts

instance Types t => Types (Qual t) where
  apply s (ps :=> t) = apply s ps :=> apply s t
  tv (ps :=> t)      = tv ps `union` tv t

instance Types Pred where
  apply s (IsIn i ts) = IsIn i (apply s ts)
  tv (IsIn i ts)      = tv ts

-- a variation that only deals with independent parameter 
tv0 :: Pred -> [Tyvar]
tv0 (IsIn i (t:_)) = tv t

tvs0 :: [Pred] -> [Tyvar]
tvs0 = nub . concat . map tv0    

instance Unify Pred where
   mgu = lift mgu

instance Match Pred where
   match = lift match

lift m (IsIn i ts) (IsIn i' ts')
         | i == i'   = m ts ts'
         | otherwise = fail "classes differ"

type Class    = ([Tyvar]{- type parameters -}, [Pred]{- super classes -},
		 [Inst]{- instances,
	                  n.b. all instance information is stored in Variant classes (for no reason). 
 		          In Record classes, this field not used. -},
		 Bool {- Variant or not  -},
		 [Id] {- methods(constructors) -}
		      {- Definitions can be extracted from [Assump]. -}  )
type Inst     = Qual (Pred{- Variant -}, Tyvar, Pred{- Record -}
		      {- Definitions should be extracted from [Assump]. -})

-----------------------------------------------------------------------------

data ClassEnv = ClassEnv { classes  :: Id -> Maybe Class{-, deleted by kagawa
                           defaults :: [Type] -} }

sig       :: ClassEnv -> Id -> [Tyvar]
sig ce i   = case classes ce i of
               Just (vs, is, its, v, ids) -> vs
               Nothing -> error ("sig: "++i) -- ɲ by kagawa

super     :: ClassEnv -> Id -> [Pred]
super ce i = case classes ce i of
               Just (vs, is, its, v, ids) -> is
               Nothing -> error ("super: "++i) -- ɲ by kagawa

insts     :: ClassEnv -> Id -> [Inst]
insts ce i = case classes ce i of 
               Just (vs, is, its, v, ids) -> its
               Nothing -> error ("inst: "++i) -- added by kagawa

methods     :: ClassEnv -> Id -> [Id]
methods ce i = case classes ce i of 
                 Just (vs, is, its, v, ids) -> ids
                 Nothing -> error ("method: "++i)


defined :: Maybe a -> Bool
defined (Just x) = True
defined Nothing  = False

modify       :: ClassEnv -> Id -> Class -> ClassEnv
modify ce i c = ce{classes = \j -> if i==j then Just c
                                           else classes ce j}

initialEnv :: ClassEnv
initialEnv  = ClassEnv { classes  = \i -> fail "class not defined"{-,
                         defaults = [tInteger, tDouble] -} }

type EnvTransformer = ClassEnv -> Maybe ClassEnv

infixr 5 <:>
(<:>)       :: EnvTransformer -> EnvTransformer -> EnvTransformer
(f <:> g) ce = do ce' <- f ce
                  g ce'

addClass                              :: Id -> [Tyvar] -> [Pred] -> [Id] -> EnvTransformer
addClass i vs ps ids ce  -- method (ids) is added to parameters
 | defined (classes ce i)              = fail "class already defined"
 | any (not . defined . classes ce . predHead) ps = fail "superclass not defined"
 | otherwise                           = return (modify ce i (vs, ps, [], False, ids))

-- ɲ by kagawa
addVariant                              :: Id -> [Tyvar] -> [Pred] -> [Id] -> EnvTransformer
addVariant i vs ps ids ce
 | defined (classes ce i)              = fail "variant already defined"
 | any (not . defined . classes ce . predHead) ps = fail "superclass not defined"
 | otherwise                           = return (modify ce i (vs, ps, [], True, ids))

-- deleted by kagawa
-- addPreludeClasses :: EnvTransformer
-- addPreludeClasses  = addCoreClasses <:> addNumClasses

atyvar = Tyvar "a" Star
atype  = TVar atyvar
asig   = [atyvar]

mtyvar = Tyvar "m" (Kfun Star Star)
mtype  = TVar mtyvar
msig   = [mtyvar]

-- deleted by kagawa
-- addCoreClasses ::   EnvTransformer
-- addCoreClasses  =   addClass "Eq" asig []
--                 <:> addClass "Ord" asig [IsIn "Eq" [atype]]
--                 <:> addClass "Show" asig []
--                 <:> addClass "Read" asig []
--                 <:> addClass "Bounded" asig []
--                 <:> addClass "Enum" asig []
--                 <:> addClass "Functor" msig []
--                 <:> addClass "Monad" msig []

-- addNumClasses  ::   EnvTransformer
-- addNumClasses   =   addClass "Num" asig [IsIn "Eq" [atype],
--                                          IsIn "Show" [atype]]
--                 <:> addClass "Real" asig [IsIn "Num" [atype],
--                                           IsIn "Ord" [atype]]
--                 <:> addClass "Fractional" asig [IsIn "Num" [atype]]
--                 <:> addClass "Integral" asig [IsIn "Real" [atype],
--                                               IsIn "Enum" [atype]]
--                 <:> addClass "RealFrac" asig [IsIn "Real" [atype],
--                                               IsIn "Fractional" [atype]]
--                 <:> addClass "Floating" asig [IsIn "Fractional" [atype]]
--                 <:> addClass "RealFloat" asig [IsIn "RealFrac" [atype],
--                                                IsIn "Floating" [atype]]

-- modified by kagawa
addInst                        :: [Pred] -> (Pred, Tyvar, Pred) -> EnvTransformer
addInst ps p@(p1@(IsIn i _), tv, p2@(IsIn j _)) ce
 | not (defined (classes ce i)) = fail "no varaint for instance"
 | not (defined (classes ce j)) = fail "no record for instance"
 | any (overlap p) qs           = fail "overlapping instance"
 | otherwise                    = return (modify ce i c)
   where its = insts ce i
         qs  = [ q | (_ :=> q) <- its ]
         c   = (sig ce i, {- observe "addInst" -} super ce i, (ps:=>(p1, tv, p2)) : its,
		variantPred ce i, methods ce i)

-- modified by kagawa
overlap       :: (Pred, Tyvar, Pred) -> (Pred, Tyvar, Pred) -> Bool
overlap (p1, _, p2) (q1, _, q2)    = defined (mgu (p1, p2) (q1, q2))

-- modified by kagawa
exampleInsts ::  EnvTransformer
exampleInsts =   {- addPreludeClasses
             <:> -} addInst [] (IsIn "Unit" [], atyvar, IsIn "Ord" [])
             <:> addInst [] (IsIn "Char" [], atyvar, IsIn "Ord" [])
             <:> addInst [] (IsIn "Int" [], atyvar, IsIn "Ord" [])
             <:> addInst [IsIn "Ord" [TVar (Tyvar "a" Star)],
                          IsIn "Ord" [TVar (Tyvar "b" Star)]]
                         (IsIn "Tuple2" [TVar (Tyvar "a" Star), TVar (Tyvar "b" Star)],
                          Tyvar "c" Star, IsIn "Ord" [])

-----------------------------------------------------------------------------

-- modified by kagawa
--  We assume that the first parameter is the placeholder.
bySuper :: ClassEnv -> Pred -> [Pred]
bySuper ce p@(IsIn i ts)
 = p : concat (map (bySuper ce) supers)
   where supers = apply s (super ce i)
         s      = zip (sig ce i) ts

{-
-- deleted by kagawa
byInst                   :: ClassEnv -> Pred -> Maybe (Subst, [Pred])
byInst ce p@(IsIn i t)    = msum [ tryInst it | it <- insts ce i ]
 where tryInst (ps :=> h) = do { let { IsIn pc (x:xs) = p;
                               	       IsIn hc (y:ys) = h } ;
                               	 if pc==hc then
                               	   do { u <- mgu y x ;  
                               		v <- mgu ys xs ;
                               		let { s = v @@ u } ;
                               		Just (v, map (apply s) ps) }
                               	 else Nothing }
-}



-- modified by kagawa
entail        :: ClassEnv -> [Pred] -> Pred -> Maybe Subst
entail = scEntail

{-
entail ce ps (p@(IsIn i (x:xs))) = loop (concat (map (bySuper ce) ps))  `mplus`
                    case byInst ce p of
                      Nothing -> Nothing
                      Just (s, qs) -> foldl foo (Just s) qs 
 where loop [] = Nothing
       loop ((IsIn j (y:ys)):qs) =
           if i==j && x==y then do s <- mgu xs ys
                                   Just s
                           else loop qs
       foo Nothing _  = Nothing
       foo (Just s) q = case entail ce ps q of
                          Just s' -> Just (s'@@s)
                          Nothing -> Nothing
-}

-----------------------------------------------------------------------------

-- modified by kagawa
simplify   :: ClassEnv -> [Pred] -> (Subst, [Pred])
simplify ce ps = let (s, rs) = simplifyA ce ps in (s, tidySuper ce rs)
{-
simplify   :: ClassEnv -> [Pred] -> (Subst, [Pred])
simplify ce ps = let ps1 = -- observe "simplify/ps1" id $
			   concat (map (bySuper ce) ps) in
                 let (s, rs) = -- observe "simplify/rs" id $
		               loop [] nullSubst ps in
	         (s, rs)		      
 where loop rs s0 []                      = (s0, rs)
       loop rs s0 (p:ps) = case scEntail1 ce (rs++ps) p of
                             Just s -> if s == nullSubst 
                          	       then loop rs s0 ps
                          	       else loop [] (s@@s0) (apply s (rs++ps))
                             Nothing -> loop (p:rs) s0 ps
-} 

simplifyA ce ps = let ps1 ={- unfold super classes -}
                           concat (map (bySuper ce) ps) in
                  loop [] nullSubst ps1	      
 where -- first loop {- unify preds Preds that have the same head -}
       loop rs s0 []                      = (s0, rs)
       loop rs s0 (p:ps) = case scEntail0 ce (rs++ps) p of
                             Just s -> if s == nullSubst 
                          	       then loop ({-p:-}rs) s0 ps
                          	       else loop [] (s@@s0) (apply s (rs++p:ps))
                             Nothing -> loop (p:rs) s0 ps


tidySuper ce rs = loop2 [] rs
  where
       -- second loop {- erace super classes -}
       loop2 rs []     = rs
       loop2 rs (p:ps) = if scEntail1 ce (rs++ps) p
			 then loop2 rs ps
                         else loop2 (p:rs) ps 

		     

reduce      :: ClassEnv -> [Pred] -> (Subst, [Pred])
reduce ce ps0 = simplify ce ps0 
                
       
{-
-- deleted by kagawa
elimTauts   :: ClassEnv -> [Pred] -> (Subst, [Pred])
elimTauts ce ps = loop [] nullSubst ps
  where loop qs s0 [] = (s0, qs)
        loop qs s0 (p:ps) = case entail ce [] p of
                           Just s -> if s == nullSubst
                                     then loop qs s0 ps
                                     else loop [] (s@@s0) (apply s (qs++ps))
                           Nothing -> loop (p:qs) s0 ps
-}

scEntail        :: ClassEnv -> [Pred] -> Pred -> Maybe Subst
scEntail ce ps (IsIn i (x:xs)) = loop (concat (map (bySuper ce) ps))
  where loop [] = Nothing
        loop ((IsIn j (y:ys)):qs) =
           if i==j && x==y then do s <- -- observe "*scEntail1"
					mgu xs ys
                                   Just s
                           else loop qs

-- used in the latter half of simplify -- returns True if there are same Preds
scEntail1        :: ClassEnv -> [Pred] -> Pred -> Bool
scEntail1 ce ps (IsIn i (x:xs)) = loop (concat (map (bySuper ce) ps))
  where loop [] = False
        loop ((IsIn j (y:ys)):qs) =
           if i==j && x==y then True else loop qs

-- used in the former part of simplify -- unify the same Preds
scEntail0        :: ClassEnv -> [Pred] -> Pred -> Maybe Subst
scEntail0 ce ps (IsIn i (x:xs)) = loop ps
  where loop [] = Nothing
        loop ((IsIn j (y:ys)):qs) =
           if i==j && x==y then do s <- -- observe "*scEntail0"
					mgu xs ys
                                   Just s
                           else loop qs

-----------------------------------------------------------------------------

-- added by kagawa
variantPred :: ClassEnv -> Id -> Bool
variantPred ce i = case classes ce i of
				Just (vs, is, its, v, ids) -> v
                                Nothing -> error ("variantPred: "++i)

recordPred :: ClassEnv -> Id -> Bool
recordPred ce i = case classes ce i of
				Just (vs, is, its, v, ids) -> not v
                                Nothing -> error ("recordPred: "++i)


findInst ce (v, r) = find (\ (_ :=> (p1, _, p2)) -> predHead p2 ==  r)
                                           (-- observe "*insts"
					    insts ce ({- observe "*predHead" -} v));

-- main check routine
-- ToDo: mixture of arity
checkPreds :: Monad m => ClassEnv -> [Pred] -> m (Subst, [Pred])
checkPreds ce ps = let pairs    = [(v, r) | v <- vs, r <- rs, phv v == phv r ]
                                   where  hd (TAp f a) = hd f
                                          hd tv        = tv  
                                          phv (IsIn _ (x:_)) = hd x  
                                          (vs, rs) = partition (\ (IsIn i _) -> variantPred ce i) ps 
                       loop [] = return (nullSubst, ps) 
                       loop ((v, r):rest) =
			   case -- observe "*checkPreds-0"
			     findInst ce (predHead v, predHead r) of
                               Nothing -> fail ("no instance for "++pretty v ++" and "++pretty r)  
                               Just (qs :=> (pv, pt, pr)) ->
                                   let IsIn vn (t:ts) = v
                                       IsIn rn (_:ss) = r in 
                                   if -- see if they match
                                      maybe False id  
                                        (do { s0 <- -- observe "match"
					            match [pv, pr] {- [v, r] -} [IsIn vn ts, IsIn rn ss];
                                              s1 <- varBind pt t;
                                       	      s  <- merge s0 s1;
                                              let { qs1 = apply s qs };
                                              Just (all (\ q -> q `elem` ps) qs1) })
                                   then loop rest
                                   else -- unify and recurse
                                     do 
                                       s  <- -- observe "*checkPreds 1"
			       		     mgu [pv, pr] {- [v, r] -}  [IsIn vn ts, IsIn rn ss]
                                       let pt1 = apply s (TVar pt)
                                           t1  = apply s t
                                       s' <- -- observe "*checkPreds 2"
			       	             mgu pt1 t1
                                       let s0 = s' @@ s
                                       let qs1 = apply s0 qs
                                       let ps1 = apply s0 ps 
                                       let (s1, ps2) = -- observe "*checkPreds-3"
					               reduce ce (qs1++ps1)
                                       (s2, ps3) <- -- observe "*checkPreds-4" 
				                    checkPreds ce ps2 -- recurse
                                       return (s2 @@ s1 @@ s0, ps3)
	           in loop pairs
                                              

