-- modified so that the type inference algorithm returns the ambiguity information and the translated expression

-----------------------------------------------------------------------------
-- TIMain:	Type Inference Algorithm
-- 
-- Part of `Typing Haskell in Haskell', version of November 23, 2000
-- Copyright (c) Mark P Jones and the Oregon Graduate Institute
-- of Science and Technology, 1999-2000
-- 
-- This program is distributed as Free Software under the terms
-- in the file "License" that is included in the distribution
-- of this software, copies of which may be obtained from:
--             http://www.cse.ogi.edu/~mpj/thih/
-- 
-----------------------------------------------------------------------------

module TIMain where
import List( (\\), intersect, union, partition)
import Id
import Kind
import Type
import Subst
import Pred
import Scheme
import Assump
import TIMonad
import Infer
import Lit
import Pat
import StaticPrelude

-- for debug
-- import Hugs.Observe

-----------------------------------------------------------------------------

data Expr = Var   Id
          | Lit   Literal
          | Const Id    -- no type info in Id -- by kagawa
          | Ap    Expr Expr
          | Let   BindGroup Expr
          | Lam   Alt
          | If    Expr Expr Expr
          | Case  Expr [(Pat,Expr)]
	    deriving Show  -- added by kagawa

-----------------------------------------------------------------------------
-- The following helper functions are used to construct sample programs; if we
-- change the representation of Expr above, then we need only change the
-- definitions of the following combinators to match, and do not need to
-- rewrite all the test code.

ap              = foldl1 Ap
evar v          = (Var v)
elit l          = (Lit l)
econst c        = (Const c)
elet e f        = foldr Let f (map toBg e)

toBg           :: [(Id, Maybe Scheme, [Alt])] -> BindGroup
toBg g          = ([(v, t, alts) | (v, Just t, alts) <- g ],
                   filter (not . null) [[(v,alts) | (v,Nothing,alts) <- g]])

pNil            = PCon (idAssump nilCfun) []
pCons x y       = PCon (idAssump consCfun) [x,y]

eNil            = econst (idAssump nilCfun)
eCons x y       = ap [ econst (idAssump consCfun), x, y ]

{-
ecase           = Case
elambda         = Lam
eif             = If
-}

ecase d as      = elet [[ ("_case",
                           Nothing,
                           [([p],e) | (p,e) <- as]) ]]
                       (ap [evar "_case", d])
eif c t f       = ecase c [(PCon (idAssump trueCfun) [], t),(PCon (idAssump falseCfun) [], f)]
elambda alt     = elet [[ ("_lambda",
                           Nothing,
                           [alt]) ]]
                             (evar "_lambda")
eguarded        = foldr (\(c,t) e -> eif c t e) efail
efail           = Const ("FAIL" {- :>: Forall [Star] ([] :=> TGen 0) -})
esign e t       = elet [[ ("_val", Just t, [([],e)]) ]] (evar "_val")

eCompFrom p e c = ap [ econst (idAssump mbindMfun), e, elambda ([p],c) ]
eCompGuard e c  = eif e c eNil
eCompLet bgs c  = elet bgs c
eListRet e      = eCons e eNil

-----------------------------------------------------------------------------

type ExtAmbiguity = ([FreezedAmbiguity], Scheme, Id) -- types with Ambiguity and id's of aux functions
type ExtType = (Type, [ExtAmbiguity], Expr) -- used as return types of ti*
                                            -- Expr is an expression that contains Coersions

theStr = "''"

--  not in Prelude?
fst3 (a, b, c) = a
snd3 (a, b, c) = b
thd3 (a, b, c) = c

-- only positions of the ambiguous types are necessary.
freezeType :: [Tyvar] -> Type -> Scheme
freezeType vs t = let vs0 = tv t
                      vs1 = (vs0 \\ vs) ++ vs
                      ks  = map kind vs1
                      s   = zip vs1 (map TGen [0..])
                  in Forall ks (apply s ([]:=>t)) 

-- modified by kagawa
tiExpr                       :: Infer Expr ExtType
tiExpr ce as (Var i)          = do sc         <- find i as
                                   (ps :=> t) <- freshInst sc
                                   return (ps, (t, [], Var i))
tiExpr ce as (Const (i{-:>:sc-})) = do sc <- find i as
                                       (ps :=> t) <- freshInst sc
                                       return (ps, (t, [], Const i))
tiExpr ce as (Lit l)          = do (ps,t) <- tiLit l
                                   return (ps, (t, [], Lit l))
tiExpr ce as (Ap e f)         = do (ps, (te, ae, ee)) <- tiExpr ce as e
                                   (qs, (tf, af, ef)) <- tiExpr ce as f
                                   t       <- newTVar Star
                                   unify (tf `fn` t) te
				   s0 <- getSubst	 
				   (s1, ps1) <- checkPreds ce (apply s0 (ps++qs))
				   let s   = s1 @@ s0 		
 			               tvs = tv (apply s t) ++ tv (apply s as)
				       amb = ambiguities ce tvs ps1
				   case amb of
 					[] -> return (ps1, (t, ae++af, Ap ee ef))
			                _  -> do
  					        fid <- newId -- insert coersion
                                                let amb1 = freezeAmbiguities amb 
                                                    tf1  = freezeType (map fst amb) (apply s tf)
					        return (ps1, (t, (amb1, tf1, fid):ae++af,
							      Ap ee (Ap (Var ("id"++theStr++fid)) ef)));
tiExpr ce as (Let bg e)       = do (ps, (as', am1, bg1)) <- tiBindGroup ce as bg
                                   (qs, (t, am2, e1))   <- tiExpr ce (as' ++ as) e
                                   return (ps ++ qs, (t, am1++am2, Let bg1 e1))
tiExpr ce as (Lam alt)
  = do (ps, (t, amb, alt')) <- tiAlt ce as alt 
       return (ps, (t, amb, Lam alt'))
tiExpr ce as (If e e1 e2)
  = do (ps,(t, am, f))   <- tiExpr ce as e
       unify t tBool
       (ps1,(t1,am1,f1)) <- tiExpr ce as e1
       (ps2,(t2,am2,f2)) <- tiExpr ce as e2
       unify t1 t2
       return (ps++ps1++ps2, (t1,am++am1++am2, If f f1 f2))
tiExpr ce as (Case e branches)
  = do (ps, (t, amb, e')) <- tiExpr ce as e
       v       <- newTVar Star
       let tiBr (pat, f)
            = do (ps, (as',t')) <- tiPat ce as pat
                 unify t t'
                 (qs, (t'',am,f1))   <- tiExpr ce (as'++as) f
                 unify v t''
                 return (ps++qs,am,(pat, f1))
       pss <- mapM tiBr branches
       return (ps++concat (map fst3 pss), (v, amb++concat (map snd3 pss), Case e' (map thd3 pss)))

-----------------------------------------------------------------------------

type Alt = ([Pat], Expr)

-- modified by kagawa
tiAlt                :: Infer Alt (Type, [ExtAmbiguity], Alt) 
tiAlt ce as (pats, e) = do (ps, (as', ts)) <- tiPats ce as pats
                           (qs,(t, amb ,e1))  <- tiExpr ce (as'++as) e
                           return (ps++qs, (foldr fn t ts, amb, (pats, e1)))

-- modified by kagawa
tiAlts             :: ClassEnv -> [Assump] -> [Alt] -> Type -> TI ([Pred], [ExtAmbiguity], [Alt])
tiAlts ce as alts t = do psts <- mapM (tiAlt ce as) alts
                         mapM (unify t) (map (fst3.snd) psts)
                         return (concat (map fst psts), concat (map (snd3.snd) psts), map (thd3.snd) psts)

-----------------------------------------------------------------------------
-- modified by kagawa
split :: Monad m => ClassEnv -> [Tyvar] -> [Tyvar] -> [Pred]
                      -> m ([Pred], [Pred])
split ce fs gs ps = do let (ds, rs) = partition (all (`elem` fs) . tv) ps
                       -- rs' <- defaultedPreds ce (fs++gs) rs 
		       -- let ds = observe "split - ds" id _ds  
		       -- let rs = observe "split - rs" id _rs  
                       let rs' = concat (map snd (ambiguities ce (fs++gs) rs)) 
		       -- let rs' = observe "split - rs'" id _rs'  
                       return (ds, rs \\ rs')

type Ambiguity       = (Tyvar, [Pred])
type FreezedAmbiguity = (Kind, [Id])  -- the kind of ambiguous variable and names of involved classes
-- use only independent variables as targets
ambiguities         :: ClassEnv -> [Tyvar] -> [Pred] -> [Ambiguity]
ambiguities ce vs ps = [ (v, filter (elem v . tv0) ps) | v <- tvs0 ps \\ vs ]

freezeAmbiguities  :: [Ambiguity] -> [(Kind, [Id])]
freezeAmbiguities ambs = map (\ (tv, ps) -> (kind tv, map (\ (IsIn i _) -> i) ps)) ambs

{-
-- deleted
numClasses :: [Id]
numClasses  = ["Num", "Integral", "Floating", "Fractional",
               "Real", "RealFloat", "RealFrac"]

stdClasses :: [Id]
stdClasses  = ["Eq", "Ord", "Show", "Read", "Bounded", "Enum", "Ix",
               "Functor", "Monad", "MonadPlus"] ++ numClasses

-- modified by kagawa
candidates           :: ClassEnv -> Ambiguity -> [Type]
candidates ce (v, qs) = [ t' | let is = [ i | IsIn i t <- qs ]
                                   ts = [ t | IsIn i t <- qs ],
                               all ([TVar v]==) ts,
                               any (`elem` numClasses) is,
                               all (`elem` stdClasses) is,
                               t' <- defaults ce,
                               loop [ IsIn i [t'] | i <- is ] ]
  where loop [] = True
        loop (p:ps) = case entail ce [] p of
                        Just _  -> loop ps  -- no dependent parameters 
                        Nothing -> False

withDefaults :: Monad m => ([Ambiguity] -> [Type] -> a)
                  -> ClassEnv -> [Tyvar] -> [Pred] -> m a
withDefaults f ce vs ps
    | any null tss  = fail "cannot resolve ambiguity"
    | otherwise     = return (f vps (map head tss))
      where vps = ambiguities ce vs ps
            tss = map (candidates ce) vps

defaultedPreds :: Monad m => ClassEnv -> [Tyvar] -> [Pred] -> m [Pred]
defaultedPreds  = withDefaults (\vps ts -> concat (map snd vps))

defaultSubst   :: Monad m => ClassEnv -> [Tyvar] -> [Pred] -> m Subst
defaultSubst    = withDefaults (\vps ts -> zip (map fst vps) ts)

-}

-----------------------------------------------------------------------------

type Expl = (Id, Scheme, [Alt])

-- modified by kagawa
tiExpl :: ClassEnv -> [Assump] -> Expl -> TI ([Pred], [ExtAmbiguity], Expl)
tiExpl ce as (i, sc, alts)
        = do (qs :=> t) <- freshInst sc
             (ps, amb, alts1)  <- tiAlts ce as alts t
             s0          <- getSubst
             let qs'     = apply s0 qs
                 (s1, ps') = reduce ce (apply s0 ps)
             (s11, ps'') <- checkPreds ce ps'
             extSubst (s11 @@ s1)
             s2 <- getSubst 
             let (s3, ps''') = loop [] nullSubst ps''
                            	 where
                            	   qs'' = apply s2 qs'
                            	   loop rs s [] = (s, rs)
                            	   loop rs s (p:ps) =
                            	     case entail ce qs'' p of
                            	       Just s1 -> if s1==nullSubst
                            			  then loop rs s ps 
                            			  else loop [] (s1@@s) (apply s1 (rs++ps))
                            	       Nothing -> loop (p:rs) s ps
             extSubst s3
             s4 <- getSubst                                 
             let t'      = apply s4 t
                 fs      = tv (apply s4 as)
                 gs      = tv t' \\ fs
                 sc'     = quantify gs (qs':=>t')
             (ds,rs)    <- split ce fs gs ps'''
             if sc /= sc' then
                 fail ("signature (" ++ show sc ++ ")too general (" ++ show sc')
               else if not (null rs) then
                 fail "context too weak"
               else
                 return (ds, amb, (i, sc, alts1))

-----------------------------------------------------------------------------

type Impl   = (Id, [Alt])

{-
-- deleted by kagawa
restricted   :: [Impl] -> Bool
restricted bs = any simple bs
 where simple (i,alts) = any (null . fst) alts
-}

-- modified by kagawa
tiImpls         :: Infer [Impl] ([Assump], [ExtAmbiguity], [Impl])
tiImpls ce as bs = do
		      -- let bs = observe "*bs" id _bs
                      ts <- mapM (\_ -> newTVar Star) bs
                      let is    = map fst bs
                          -- is = observe "is" id _is
                          scs   = map toScheme ts
                          as'   = zipWith (:>:) is scs ++ as
                          altss = map snd bs
                      paess <- sequence (zipWith (tiAlts ce as') altss ts)
                      s0   <- getSubst
                      let pss = map fst3 paess
			  ams = concat (map snd3 paess)
                          altss1 = map thd3 paess
		          (s1, ps') = {- observe "reduce" -}
				      reduce ce (apply s0 (concat pss))
                      (s2, ps'') <- {-- observe "checkPreds" --}
				    checkPreds ce ps'
		      -- let ps'' = observe "ps''" id _ps''
                      extSubst (s2 @@ s1)
                      s <- getSubst
                      let ts'     = -- observe "*apply"
			            apply s ts
			  -- ts' = observe "*ts'" id _ts'
                          fs      = tv (apply s as)
                          vss     = -- observe "*map"
			            map tv ts'
			  -- vss = observe "*vss" id _vss

                      (ds,rs) <- -- observe "*split" 
				 split ce fs (foldr1 intersect vss) ps''
		      let vss2 = map tv rs 		       
                          gs   = foldr1 union (vss++vss2) \\ fs
			  -- gs = observe "gs" id _gs	     

		      -- let rs = observe "*rs" id _rs		       
--                       if -- observe "restricted"
-- 			 restricted bs then
--                           let -- gs0 = observe "id"  id gs
-- 			      -- rs0 = observe "rs0" id rs
			     
-- 			      gs'  = gs \\ {- - observe "tv" - -} tv rs
-- 			      -- gs' = observe "identity" id _gs'
--                               _scs' = map ({- observe "quantify" -}
-- 					  quantify gs' . ([]:=>)) ts'
-- 			      scs' = observe "scs'" id _scs'
                              
--                           in {- - observe "return" - -}
-- 			     return (ds++rs, -- observe "zipWith0 "
-- 				             zipWith (:>:) is scs')
--                         else
                      let scs' = map (quantify gs . (rs:=>)) ts' {- in -}
                      return (ds, -- observe "*zipWith"
			               (zipWith (:>:) is scs', ams, zip is altss1))

-----------------------------------------------------------------------------

type BindGroup  = ([Expl], [[Impl]])

-- modified by kagawa
tiBindGroup :: Infer BindGroup ([Assump], [ExtAmbiguity], BindGroup)
tiBindGroup ce as (es,iss) =
  do let as' = [ v:>:sc | (v,sc,alts) <- es ]
     (ps, (as'', amb'', iss')) <- tiSeq ({- -observe "impls" - -}tiImpls) ce (as'++as) iss
     -- let ps = observe "ps" ps00
     qaess        <- mapM ({- observe "expl" -} tiExpl ce (as''++as'++as)) es
     let qss = map fst3 qaess
         amb1 = concat (map snd3 qaess)
         es'  = map thd3 qaess
     return (ps++concat qss, (as''++as', amb''++amb1, (es',iss')))

-- modified by kagawa
tiSeq                  :: Infer bg ([Assump], [ExtAmbiguity], bg) -> Infer [bg] ([Assump], [ExtAmbiguity], [bg])
tiSeq ti ce as []       = return ([],([], [], []))
tiSeq ti ce as (bs:bss) = do (ps,(as', am', bs'))  <- ti ce as bs
                             (qs,(as'',am'', bss')) <- tiSeq ti ce (as'++as) bss
                             return (ps++qs, (as''++as', am''++am', bs':bss'))
-----------------------------------------------------------------------------
